#!/bin/bash
################################################################################
# Copyright (c) 2013, 2015 Wind River Systems, Inc.
#
# The right to copy, distribute, modify, or otherwise make use of this
# software may be licensed only pursuant to the terms of an applicable Wind
# River license agreement.
#
################################################################################

SYSFS_DEVICE_PATH="/sys/bus/pci/devices"

################################################################################
# Utility function to print the status of a command result
#
################################################################################
function print_status()
{
    if [ "$1" -eq "0" ]; then
        echo "[  OK  ]"
    else
        echo "[FAILED]"
    fi
}

################################################################################
# Utility function to expand a sequence of numbers (e.g., 0-7,16-23)
################################################################################
function expand_sequence
{
    SEQUENCE=(${1//,/ })
    DELIMITER=${2:-","}

    LIST=
    for entry in ${SEQUENCE[@]}
    do
        range=(${entry/-/ })
        a=${range[0]}
        b=${range[1]:-${range[0]}}

        for i in $(seq $a $b)
        do
            LIST="${LIST}${DELIMITER}${i}"
        done
    done
    echo ${LIST:1}
}


################################################################################
# Converts a CPULIST (e.g., 0-7,16-23) to a CPUMAP (e.g., 0x00FF00FF).  The
# CPU map is returned as a string representation of a large hexidecimal
# number but without the leading "0x" characters.
#
################################################################################
function cpulist_to_cpumap
{
    local CPULIST=$1
    local NR_CPUS=$2
    local CPUMAP=0
    local CPUID=0

    for CPUID in $(expand_sequence $CPULIST " ")
    do
        if [ "${CPUID}" -lt "${NR_CPUS}" ]; then
            CPUMAP=$(echo "${CPUMAP} + (2^${CPUID})" | bc -l)
        fi
    done

    echo "obase=16;ibase=10;${CPUMAP}" | bc -l
    return 0
}

################################################################################
# Converts a CPUMAP (e.g., 0x00FF00FF) to a CPULIST (e.g., 0-7,16-23).  The
# CPUMAP is expected in hexidecimal (base=10) form without the leading "0x"
# characters.
#
################################################################################
function cpumap_to_cpulist
{
    local CPUMAP=$(echo "obase=10;ibase=16;$1" | bc -l)
    local NR_CPUS=$2
    local COUNT=1
    local CPUID=0
    local LAST=9999

    for CPUID in $(seq 0 $((NR_CPUS - 1)));
    do
        ## Since 'bc' does not support any bitwise operators this expression:
        ##     if (CPUMAP & (1 << CPUID))
        ## has to be rewritten like this:
        ##     if (CPUMAP % (2**(CPUID+1)) > ((2**(CPUID)) - 1))
        ##
        ISSET=$(echo "scale=0; (${CPUMAP} % 2^(${CPUID}+1)) > (2^${CPUID})-1" | bc -l)
        if [ "${ISSET}" -ne 1 ]; then
            continue
        fi

        if [ ${CPUID} -eq $((LAST+1)) ]; then
            COUNT=$((COUNT+1))
            LAST=$CPUID
            continue
        elif [ ${LAST} -ne 9999 -a ${COUNT} -le 1 ]; then
            printf ","
        fi

        if [ ${COUNT} -gt 1 ]; then
            printf -- "-%u," ${LAST}
        fi

        printf "%u" $CPUID
        LAST=$CPUID
        COUNT=1
    done

    if [ ${COUNT} -gt 1 ]; then
        printf -- "-%u" ${LAST}
    fi

    printf "\n"
    return 0
}


################################################################################
# Convert a list of CPU values to a string representation compatible with the
# DPDK EAL --lcores argument.  Since we want to use lcoreid=0 for the
# master/mgmt core and we could also end up running a fastpath engine on cpu0
# we need to use explicit lcoreid values for each cpu to avoid a conflict
# between both threads wanting to be lcoreid=0.  The master/mgmt thread will
# always be lcoreid=0 and the fastpath engines will always be assigned from
# lcoreid=1 regardless of which cpu they are actually running on.
#
################################################################################
function cpulist_to_lcores
{
    local CPULIST=$1
    local LCOREID=${2:-"1"}
    local LCORES=""
    local SEPARATOR=""

    for CPUID in $(expand_sequence ${CPULIST} " "); do
        LCORES="${LCORES}${SEPARATOR}${LCOREID}@${CPUID}"
        LCOREID=$((LCOREID+1))
        SEPARATOR=","
    done

    echo $LCORES
}

################################################################################
# Bitwise NOT of a hexidecimal representation of a CPULIST.   The value is
# returned as a hexidecimal value but without the leading "0x" characters
#
################################################################################
function invert_cpumap
{
    local CPUMAP=$(echo "obase=10;ibase=16;$1" | bc -l)
    local NR_CPUS=$2
    local INVERSE_CPUMAP=0

    for CPUID in $(seq 0 $((NR_CPUS - 1)));
    do
        ## See comment in previous function
        ISSET=$(echo "scale=0; (${CPUMAP} % 2^(${CPUID}+1)) > (2^${CPUID})-1" | bc -l)
        if [ "${ISSET}" -eq 1 ]; then
            continue
        fi

        INVERSE_CPUMAP=$(echo "${INVERSE_CPUMAP} + (2^${CPUID})" | bc -l)
    done

    echo "obase=16;ibase=10;${INVERSE_CPUMAP}" | bc -l
    return 0
}

################################################################################
# Builds the complement representation of a CPULIST
#
################################################################################
function invert_cpulist
{
    local CPULIST=$1
    local NR_CPUS=$2
    local CPUMAP=$(cpulist_to_cpumap ${CPULIST} ${NR_CPUS})
    cpumap_to_cpulist $(invert_cpumap ${CPUMAP} ${NR_CPUS}) ${NR_CPUS}
    return 0
}

################################################################################
# Retrieve the number of CPUs
#
################################################################################
function cpu_count
{
    echo `grep -c ^processor /proc/cpuinfo`
    return 0
}

################################################################################
# Retrieve the number of NUMA nodes
#
################################################################################
function numa_count
{
    echo `ls -d /sys/devices/system/node/node* 2>/dev/null | wc -l`
    return 0
}

################################################################################
# Retrieve the current system product name
#
# Note: comment lines are stripped from the dmidecode command to avoid
# failures when dmidecode reports warnings about BIOS versions that are newer
# than what it can understand.
#
################################################################################
function product_name
{
    echo `dmidecode --string 'system-product-name' | grep -vE "^#"`
    return 0
}

################################################################################
# Log if debug is enabled via VSWITCH_DEBUG
#
################################################################################
function log_debug
{
    if [ ! -z "${VSWITCH_DEBUG}" ]; then
        logger -t $0 -p daemon.debug $@
    fi
}

################################################################################
# Log unconditionally to STDERR
#
################################################################################
function log_error
{
    logger -t $0 -p daemon.error $@
}

################################################################################
# Log unconditionally to STDOUT
#
################################################################################
function log
{
    logger -t $0 -p daemon.info $@
}


################################################################################
# Wait for a process to stop running.
#
################################################################################
function wait_for_proc_stop()
{
    PROGNAME=$1
    TIMEOUT=${2:-"5"}

    for I in $(seq 1 $TIMEOUT); do
        PID=$(pidof $PROGNAME 2> /dev/null)
        if [ $? -ne 0 ]; then
            ## already dead
            return 0
        fi
        sleep 1
    done

    return 1
}

################################################################################
# Wait for a process to start running.
#
################################################################################
function wait_for_proc_start()
{
    PROGNAME=$1
    TIMEOUT=${2:-"5"}

    for I in $(seq 1 $TIMEOUT); do
        PID=$(pidof $PROGNAME 2> /dev/null)
        if [ $? -eq 0 ]; then
            ## already running
            return 0
        fi
        sleep 1
    done

    return 1
}

################################################################################
# obtain a list of PCI devices of a particular vendor and device id
################################################################################
function pci_device_list()
{
    local VENDORID=$1
    local DEVICEID=$2
    local SUBDEVICEID=$3
    local MAXCOUNT=$4
    local DEVICES=()
    local COUNT=0

    for PCIDEV in $(ls ${SYSFS_DEVICE_PATH});
    do
        VENDOR=$(cat ${SYSFS_DEVICE_PATH}/${PCIDEV}/vendor)
        DEVICE=$(cat ${SYSFS_DEVICE_PATH}/${PCIDEV}/device)
        SUBDEVICE=$(cat ${SYSFS_DEVICE_PATH}/${PCIDEV}/subsystem_device)
        if [ "${VENDOR}" == "${VENDORID}" -a "${DEVICE}" == "${DEVICEID}" ];
        then
            if [ "${SUBDEVICEID}" == "0x0000" -o "${SUBDEVICEID}" == "${SUBDEVICE}" ];
            then
                DEVICES+=("${PCIDEV}")
            fi
        fi
    done

    COUNT=${#DEVICES[@]}
    if [ ${MAXCOUNT} -ne 0 -a ${COUNT} -gt ${MAXCOUNT} ];
    then
        ## Truncate the list to the upper limit provided
        DEVICES=("${DEVICES[@]:$((COUNT-${MAXCOUNT})):${MAXCOUNT}}")
    fi

    echo ${DEVICES[@]}
    return 0;
}
